<!DOCTYPE html>
<html>
<head>
	<meta charset="utf-8" />
	<title>IMS Global Open Badges 2.0 Validator</title>
	<link rel="stylesheet" type="text/css" href="/static/screen.css">
    <link rel="stylesheet" href="https://maxcdn.bootstrapcdn.com/bootstrap/3.3.7/css/bootstrap.min.css" integrity="sha384-BVYiiSIFeK1dGmJRAkycuHAHRg32OmUcww7on3RYdg4Va+PmSTsz/K68vbdEjh4u" crossorigin="anonymous">
</head>

<body>
<main class="container">
<header class="row">
    <div class="col-md-8"></div>
    <div class="col-md-4">
        <a id="logo" href="https://www.imsglobal.org">
            <img src="https://www.imsglobal.org/sites/default/files/imslogo.png" alt="IMS Global Homepage">
        </a>
    </div>
</header>

<h1>IMS Global Open Badges 2.0 Validator</h1>

{% block main %}{% endblock %}

<footer>
	<h3 id="developers">Developers</h3>
	<p>Full documentation, including APIs, is <a href="https://github.com/IMSGlobal/openbadges-validator-core/blob/develop/README.md">available</a>.</p>
<h3 id="disclaimer">Disclaimer</h3>
    <p>
        IMS Global takes no responsibility for and will not be liable for this service being temporarily unavailable
        due to technical issues (or otherwise), or for any loss or damage suffered as a result of the use of or access
        to, or inability to use or access this service whatsoever.
    </p>
</footer>

</main>
</body>
</html>
