"""
INPUT Actions:
Store and process user input via the API
"""
SET_INPUT_TYPE = 'SET_INPUT_TYPE'
SET_PROFILE_ID = 'SET_PROFILE_ID'
STORE_INPUT = 'STORE_INPUT'
STORE_ORIGINAL_RESOURCE = 'STORE_ORIGINAL_JSON'

"""
GRAPH Actions:
Manage the state of the known entities related to the validation subject.
"""
ADD_NODE = 'ADD_NODE'
PATCH_NODE = 'PATCH_NODE'
UPDATE_NODE = 'UPDATE_NODE'
PATCH_NODE_REFERENCE = 'PATCH_NODE_REFERENCE'

"""
TASK Actions:
Add, update, and complete tasks.
"""
ADD_TASK = 'ADD_TASK'
UPDATE_TASK = 'UPDATE_TASK'
DELETE_TASK = 'DELETE_TASK'
DELETE_OUTDATED_NODE_TASKS = 'DELETE_OUTDATED_NODE_TASKS'
REPORT_MESSAGE = 'REPORT_MESSAGE'
RESOLVE_TASK = 'RESOLVE_TASK'
TRIGGER_CONDITION = 'TRIGGER_CONDITION'


"""
VALIDATION REPORT Actions:
Note data that should be stored in the validation report.
"""
RUN_VALIDATION_REPORT = 'RUN_VALIDATION_REPORT'
SET_OPENBADGES_VERSION = 'SET_OPENBADGES_VERSION'
SET_VALIDATION_SUBJECT = 'SET_VALIDATION_SUBJECT'
SET_VERIFIED_PROFILE = 'SET_VERIFIED_PROFILE'
