body {
	margin: 5%;
	font-family: sans-serif;
	color:rgb(80,80,80);
	background-color: rgb(255,255,255);
}

* {
    box-sizing: border-box;
/*	margin-top:5% */
}
.container {
	max-width: 1024px;
}

h1 {
	margin-top: 6em;
	margin-bottom:5%;
	text-align: center;
}

#notice {
	color: red;
}

form {
	padding:5% 5% 5% 5%;
}

form#submit {
	margin-top:3em;
	text-align: left;
	border: 1px solid rgb(190,190,190);
}

#results textarea {
	font-family: monospace;
	width: 100%;
	height: auto;
	border: 1px solid rgb(190,190,190);
}

.badgeform .badgeforminput {
    width: 100%;
}

footer {
	font-size:80%;
}

footer p {
	font-style:italic;
}

